import json
import time
from typing import Annotated

import anyio
import typer
from fastmcp import Client
from mcp.types import TextContent

from daydream import mcp
from daydream.clilib.client import app as client_app
from daydream.clilib.options import PROFILE_OPTION
from daydream.config import load_config
from daydream.config.utils import get_config_dir
from daydream.knowledge import Graph
from daydream.mcp import Context, build_mcp_server
from daydream.plugins import PluginManager
from daydream.plugins.mixins import KnowledgeGraphMixin

app = typer.Typer(context_settings={"help_option_names": ["-h", "--help"]}, no_args_is_help=True)
tools_app = typer.Typer(context_settings={"help_option_names": ["-h", "--help"]})
app.add_typer(tools_app, name="tools")
app.add_typer(client_app, name="client")


@app.command()
def start(
    profile: str = PROFILE_OPTION,
    disable_sse: bool = typer.Option(
        False, "--disable-sse", help="Disable the SSE transport for the MCP Server"
    ),
    disable_stdio: bool = typer.Option(
        False, "--disable-stdio", help="Disable the stdio transport for the MCP Server"
    ),
) -> None:
    """Start the Daydream MCP Server"""
    anyio.run(mcp.start, profile, disable_sse, disable_stdio)


@app.command()
def build_graph(
    profile: str = PROFILE_OPTION,
) -> None:
    """Build a knowledge graph for your cloud infrastructure"""

    async def _build_graph() -> None:
        print("Building graph...")

        start_time = time.perf_counter()

        graph = Graph()
        config = load_config(profile, create=True)
        output_path = (get_config_dir(profile) / "graph.json").resolve()
        plugin_manager = PluginManager(config)

        async with anyio.create_task_group() as tg:
            for plugin in plugin_manager.get_plugins_with_capability(KnowledgeGraphMixin):
                print(f"Populating graph with the {plugin.name} plugin...")
                tg.start_soon(plugin.populate_graph, graph)

        await graph.infer_edges()

        print(f"Saving graph to {output_path!s}...")
        await graph.save(output_path)

        end_time = time.perf_counter()
        print(f"Graph built in {end_time - start_time:.2f} seconds")
        print(f"Graph saved to {output_path!s}")

    anyio.run(_build_graph)


@app.command()
def visualize(
    profile: str = PROFILE_OPTION,
    topology: bool = typer.Option(False, "--topology"),
) -> None:
    """Visualize the knowledge graph topology"""

    async def _visualize() -> None:
        graph = Graph(get_config_dir(profile) / "graph.json")

        if topology:
            graph = await graph.get_topology()

        print(await graph.to_pydot())

    anyio.run(_visualize)


@tools_app.command("list")
def list_tools(
    profile: str = PROFILE_OPTION,
) -> None:
    """List all MCP tools available from enabled plugins."""

    async def _list_tools() -> None:
        config = load_config(profile)
        plugins = PluginManager(config=config)
        context = Context(
            profile=profile,
            config=config,
            plugins=plugins,
            graph=Graph(get_config_dir(profile) / "graph.json"),
        )
        mcp = build_mcp_server(context)

        tools = await mcp.get_tools()
        for key, tool in tools.items():
            cmd = [key]
            for arg, arg_data in tool.parameters["properties"].items():
                arg_type = arg_data.get("type", "unknown")
                if "anyOf" in arg_data:
                    arg_type = "|".join(t["type"] for t in arg_data["anyOf"])
                cmd.append(f"<{arg}:{arg_type}>")
            print(" ".join(cmd))

    anyio.run(_list_tools)


@tools_app.command()
def call(
    tool: str,
    arguments: Annotated[list[str] | None, typer.Argument()] = None,
    profile: str = PROFILE_OPTION,
) -> None:
    """Call an MCP tool from the command line."""

    async def _call_tool() -> None:
        config = load_config(profile)
        plugins = PluginManager(config=config)
        context = Context(
            profile=profile,
            config=config,
            plugins=plugins,
            graph=Graph(get_config_dir(profile) / "graph.json"),
        )
        mcp = build_mcp_server(context)

        tools = await mcp.get_tools()
        if tool not in tools:
            raise typer.Abort(f"Tool {tool} not found")

        tool_def = tools[tool]
        tool_args = {}
        if arguments:
            # Convert a list of arguments into a dict
            for k, v in zip(tool_def.parameters["properties"].keys(), arguments, strict=True):
                # Try decoding as JSON to handle complex arguments (lists, dicts, etc.)
                try:
                    tool_args[k] = json.loads(v)
                except json.JSONDecodeError:
                    tool_args[k] = v

        client = Client(mcp)
        async with client:
            result = await client.call_tool(tool, tool_args)
            try:
                print(next(r.text for r in result if isinstance(r, TextContent)))
            except StopIteration:
                print(f"{tool} returned no printable results")

    anyio.run(_call_tool)


def main() -> None:
    app()


if __name__ == "__main__":
    main()



from collections.abc import Iterator
from typing import TYPE_CHECKING, Any, Protocol, TypeVar, cast

from daydream.config.utils import Config, PluginConfig, PluginSettings
from daydream.utils import classproperty

if TYPE_CHECKING:
    from daydream.mcp import Context

REGISTRY: dict[str, type["Plugin"]] = {}


class PluginProtocol(Protocol):
    """Protocol for the plugins interface."""

    context: "Context"

    @classproperty
    def name(cls) -> str:
        """Converts a SomePlugin class to a "some" string, which is the key used in the configuration file.

        For example, AwsPlugin becomes "aws".
        """
        return cast("type[PluginProtocol]", cls).__name__.removesuffix("Plugin").lower()


class Plugin(PluginProtocol):
    """Base class for all plugins."""

    def __init__(self, **settings: Any) -> None:
        self._settings = settings

    def __init_subclass__(cls, **kwargs: Any) -> None:
        REGISTRY[cls.name] = cls

    @classproperty
    def default_plugin_settings(cls) -> PluginSettings:
        return {}


class PluginCapability(PluginProtocol):
    """Protocol for capabilities that plugins can have."""


HasPluginCapability = TypeVar("HasPluginCapability", bound=PluginCapability)


class PluginManager:
    """A manager for plugins and their configuration."""

    def __init__(self, config: Config) -> None:
        self._config = config
        self._plugins = {}

    def get_plugin_class(self, name: str) -> type[Plugin]:
        """Return the plugin class for a given name."""
        return REGISTRY[name]

    def get_plugin(self, name: str) -> Plugin:
        """Return the configured plugin for a given name."""
        if name not in self._plugins:
            self._plugins[name] = self.get_plugin_class(name)(**self._config.plugins[name].settings)
        return self._plugins[name]

    def get_enabled_plugins(self) -> list[Plugin]:
        """Return a list of enabled plugins."""
        return [
            self.get_plugin(name) for name, config in self._config.plugins.items() if config.enabled
        ]

    def get_plugins_with_capability(
        self, capability: type[HasPluginCapability]
    ) -> list[HasPluginCapability]:
        """Return a list of plugins that have the given capability."""
        return [plugin for plugin in self.get_enabled_plugins() if isinstance(plugin, capability)]

    def __iter__(self) -> Iterator[Plugin]:
        """Iterate over all enabled plugins."""
        return iter(self.get_enabled_plugins())

    @classproperty
    def default_config(cls) -> Config:
        cfg = Config()
        for p, plugin_cls in REGISTRY.items():
            cfg.plugins[p] = PluginConfig(enabled=True, settings=plugin_cls.default_plugin_settings)
        return cfg



from collections.abc import AsyncGenerator
from typing import Any, Literal

import anyio
from aioboto3 import Session

from daydream.knowledge import Graph
from daydream.plugins import Plugin
from daydream.plugins.aws.nodes.aws_alb_target_group import AwsAlbTargetGroup
from daydream.plugins.aws.nodes.aws_application_load_balancer import (
    AwsApplicationLoadBalancer,
)
from daydream.plugins.aws.nodes.aws_classic_load_balancer import AwsClassicLoadBalancer
from daydream.plugins.aws.nodes.aws_ebs_volume import AwsEbsVolume
from daydream.plugins.aws.nodes.aws_ec2_instance import AwsEc2Instance
from daydream.plugins.aws.nodes.aws_rds_database import AwsRdsDatabase
from daydream.plugins.aws.utils import (
    list_accessible_regions_for_service,
    swallow_boto_client_access_errors,
)
from daydream.plugins.mixins import KnowledgeGraphMixin
from daydream.utils import print


class AwsPlugin(Plugin, KnowledgeGraphMixin):
    def __init__(self, **settings: Any) -> None:
        super().__init__(**settings)
        self._session = Session()

    @property
    def session(self) -> Session:
        return self._session

    async def _ensure_session(self) -> None:
        async with self.session.client("sts") as client:
            try:
                await client.get_caller_identity()
            except Exception as e:
                print(f"Error using session to call STS GetCallerIdentity: {e}")
                raise

    async def populate_graph(self, graph: Graph) -> None:
        await self._ensure_session()
        async with anyio.create_task_group() as tg:
            tg.start_soon(self.populate_rds_databases, graph)
            tg.start_soon(self.populate_ec2_instances, graph)
            tg.start_soon(self.populate_classic_load_balancers, graph)
            tg.start_soon(self.populate_application_load_balancers, graph)
            tg.start_soon(self.populate_alb_target_groups, graph)
            tg.start_soon(self.populate_ebs_volumes, graph)

    async def populate_rds_databases(self, graph: Graph) -> None:
        async with anyio.create_task_group() as tg:
            for region in await list_accessible_regions_for_service(self.session, "rds"):
                tg.start_soon(self._populate_rds_databases_in_region, graph, region)

    async def _populate_rds_databases_in_region(self, graph: Graph, region: str) -> None:
        print(f"Populating RDS databases from {region}")

        rds_database_count = 0
        async for page in self._paginate("rds", "describe_db_instances", region):
            for instance in page["DBInstances"]:
                await graph.add_node(
                    AwsRdsDatabase(
                        node_id=instance["DBInstanceArn"],
                        raw_data=instance,
                        _graph=graph,
                    )
                )
                rds_database_count += 1

        print(f"Initialized {rds_database_count} RDS databases for {region}")

    async def populate_ec2_instances(self, graph: Graph) -> None:
        async with anyio.create_task_group() as tg:
            for region in await list_accessible_regions_for_service(self.session, "ec2"):
                tg.start_soon(self._populate_ec2_instances_in_region, graph, region)

    async def _populate_ec2_instances_in_region(self, graph: Graph, region: str) -> None:
        print(f"Populating EC2 instances from {region}")

        ec2_instance_count = 0
        async for page in self._paginate("ec2", "describe_instances", region):
            for reservation in page["Reservations"]:
                for instance in reservation["Instances"]:
                    await graph.add_node(
                        AwsEc2Instance(
                            node_id=instance["InstanceId"],
                            raw_data=instance,
                            _graph=graph,
                        )
                    )
                    ec2_instance_count += 1

        print(f"Initialized {ec2_instance_count} EC2 instances for {region}")

    async def populate_classic_load_balancers(self, graph: Graph) -> None:
        async with anyio.create_task_group() as tg:
            for region in await list_accessible_regions_for_service(self.session, "elb"):
                tg.start_soon(self._populate_classic_load_balancers_in_region, graph, region)

    async def _populate_classic_load_balancers_in_region(self, graph: Graph, region: str) -> None:
        print(f"Populating classic load balancers from {region}")

        elb_count = 0
        async for page in self._paginate("elb", "describe_load_balancers", region):
            for balancer in page["LoadBalancerDescriptions"]:
                await graph.add_node(
                    AwsClassicLoadBalancer(
                        node_id=balancer["LoadBalancerName"],
                        raw_data=balancer,
                        _graph=graph,
                    )
                )
                elb_count += 1

        print(f"Initialized {elb_count} classic load balancers for {region}")

    async def populate_application_load_balancers(self, graph: Graph) -> None:
        async with anyio.create_task_group() as tg:
            for region in await list_accessible_regions_for_service(self.session, "elbv2"):
                tg.start_soon(self._populate_application_load_balancers_in_region, graph, region)

    async def _populate_application_load_balancers_in_region(
        self, graph: Graph, region: str
    ) -> None:
        print(f"Populating application load balancers from {region}")

        alb_count = 0
        async for page in self._paginate("elbv2", "describe_load_balancers", region):
            for balancer in page["LoadBalancers"]:
                await graph.add_node(
                    AwsApplicationLoadBalancer(
                        node_id=balancer["LoadBalancerArn"],
                        raw_data=balancer,
                        _graph=graph,
                    )
                )
                alb_count += 1

        print(f"Initialized {alb_count} application load balancers for {region}")

    async def populate_alb_target_groups(self, graph: Graph) -> None:
        async with anyio.create_task_group() as tg:
            for region in await list_accessible_regions_for_service(self.session, "elbv2"):
                tg.start_soon(self._populate_alb_target_groups_in_region, graph, region)

    async def _populate_alb_target_groups_in_region(self, graph: Graph, region: str) -> None:
        print(f"Populating ALB target groups from {region}")

        alb_target_group_count = 0
        async for page in self._paginate("elbv2", "describe_target_groups", region):
            for target_group in page["TargetGroups"]:
                await graph.add_node(
                    AwsAlbTargetGroup(
                        node_id=target_group["TargetGroupArn"],
                        raw_data=target_group,
                        _graph=graph,
                    )
                )
                alb_target_group_count += 1

        print(f"Initialized {alb_target_group_count} ALB target groups for {region}")

    async def populate_ebs_volumes(self, graph: Graph) -> None:
        async with anyio.create_task_group() as tg:
            for region in await list_accessible_regions_for_service(self.session, "ec2"):
                tg.start_soon(self._populate_ebs_volumes_in_region, graph, region)

    async def _populate_ebs_volumes_in_region(self, graph: Graph, region: str) -> None:
        print(f"Populating EBS volumes from {region}")

        ebs_volume_count = 0
        async for page in self._paginate("ec2", "describe_volumes", region):
            for volume in page["Volumes"]:
                await graph.add_node(
                    AwsEbsVolume(
                        node_id=volume["VolumeId"],
                        raw_data=volume,
                        _graph=graph,
                    )
                )
                ebs_volume_count += 1

        print(f"Initialized {ebs_volume_count} EBS volumes for {region}")

    async def _paginate(
        self,
        service_name: Literal["ec2", "elb", "elbv2", "rds"],
        action: str,
        region: str,
    ) -> AsyncGenerator[dict, None]:
        async with (
            swallow_boto_client_access_errors(service_name=service_name, region=region),
            self.session.client(service_name, region_name=region) as client,
        ):
            paginator = client.get_paginator(action)
            async for page in paginator.paginate():
                yield page



import typer

import daydream.clilib.client_configure as client_configure
import daydream.clilib.client_logs as client_logs

app = typer.Typer(help="Work with clients of the Daydream MCP Server", no_args_is_help=True)
app.add_typer(client_configure.app, name="configure")
app.add_typer(client_logs.app, name="logs")



import rich
import typer

app = typer.Typer(
    help="Show logs for a client of the Daydream MCP Server, like Claude Desktop",
    no_args_is_help=True,
)


@app.command()
def claude() -> None:
    """Show logs for Claude Desktop Daydream MCP Server"""
    rich.print("Logs are available at:")
    rich.print("~/Library/Logs/Claude/mcp.log")
    rich.print("~/Library/Logs/Claude/mcp-server-daydream.log")



import json
import shlex
import shutil
from datetime import datetime
from pathlib import Path
from typing import Annotated, Any

import boto3.session
import questionary
import rich
import typer

from daydream.clilib.options import DEFAULT_PROFILE, PROFILE_OPTION
from daydream.config.utils import Config, get_config_path, save_config
from daydream.plugins.base import PluginManager

app = typer.Typer(
    help="Configure clients of the Daydream MCP Server, like Claude Desktop", no_args_is_help=True
)


def _resolve_default_claude_mcp_config_filename() -> Path:
    claude_mcp_file = (
        Path.home() / "Library" / "Application Support" / "Claude" / "claude_desktop_config.json"
    )
    return claude_mcp_file.absolute()


_default_claude_mcp_config_filename = _resolve_default_claude_mcp_config_filename()


def _resolve_default_uv() -> Path:
    uv_path = shutil.which("uv")
    if uv_path:
        return Path(uv_path).absolute()
    return Path("uv")


_default_uv_path = _resolve_default_uv()


def _resolve_default_repo_path() -> Path:
    return Path.cwd().absolute()


_default_repo_path = _resolve_default_repo_path()


def _resolve_default_claude_app() -> Path:
    return Path("/Applications/Claude.app").resolve()


_default_claude_app = _resolve_default_claude_app()


@app.command()
def claude(
    config_filename: Annotated[
        Path,
        typer.Option(
            ...,
            "--config-filename",
            help="Path to the Claude Desktop MCP config json file",
            show_default=True,
            file_okay=True,
            dir_okay=False,
            readable=True,
            writable=True,
            resolve_path=False,
            exists=False,
        ),
    ] = _default_claude_mcp_config_filename,
    repo_path: Annotated[
        Path,
        typer.Option(
            ...,
            "--repo-path",
            help="Path to the Daydream repo",
            show_default=True,
            file_okay=False,
            dir_okay=True,
            readable=True,
            writable=False,
            resolve_path=False,
            exists=True,
        ),
    ] = _default_repo_path,
    uv_path: Annotated[
        Path,
        typer.Option(
            ...,
            "--uv-path",
            help="Path to the uv executable",
            show_default=True,
            file_okay=True,
            dir_okay=False,
            readable=True,
            writable=False,
            resolve_path=False,
            exists=True,
        ),
    ] = _default_uv_path,
    claude_app: Annotated[
        Path,
        typer.Option(
            ...,
            "--claude-app",
            help="Path to the Claude Desktop app",
            show_default=True,
            file_okay=False,
            dir_okay=True,
            readable=True,
            writable=False,
            resolve_path=False,
            exists=False,
        ),
    ] = _default_claude_app,
    profile: str = PROFILE_OPTION,
    aws_profile: Annotated[
        list[str] | None,
        typer.Option(
            ...,
            "--aws-profile",
            help="One or more AWS profiles to use for the Daydream MCP Server",
        ),
    ] = None,
    enable_plugin: Annotated[
        list[str] | None,
        typer.Option(
            ...,
            "--enable-plugin",
            help="One or more plugins to enable for the Daydream MCP Server",
        ),
    ] = None,
    command_prefix: Annotated[
        str | None,
        typer.Option(
            ...,
            "--command-prefix",
            help="Command prefix to use for the Daydream MCP Server (e.g., 'aws-vault exec your_aws_profile')",
        ),
    ] = None,
    overwrite_existing: Annotated[
        bool,
        typer.Option(
            ...,
            "--overwrite-existing",
            help="Overwrite existing Daydream MCP Server config if it exists",
        ),
    ] = False,
    skip_plugin_selection: Annotated[
        bool,
        typer.Option(
            ...,
            "--skip-plugin-selection",
            help="Skip the interactive plugin selection prompt (use --enable-plugin to enable plugins)",
        ),
    ] = False,
) -> None:
    """Configure the Claude Desktop to use the Daydream MCP Server"""
    if aws_profile is None:
        aws_profile = []
    command, *args = [
        *shlex.split(command_prefix or ""),
        str(uv_path),
        "run",
        "--directory",
        str(repo_path),
        "daydream",
        "start",
        "--disable-sse",
    ]
    env = {
        "HOME": str(Path.home()),
    }
    daydream_server_config = {
        "command": shutil.which(command),
        "args": args,
        "env": env,
    }
    if not claude_app.exists():
        raise FileNotFoundError(
            f"Claude app doesn't exist at [bright_black]{claude_app}[/bright_black]. Make sure you have it installed and use --claude-app to specify a non-default location."
        )
    if profile != DEFAULT_PROFILE:
        daydream_server_config["args"].extend(["--profile", profile])
    mcp_server_config = _ensure_config_dir_and_backup_existing(config_filename, overwrite_existing)
    cfg = PluginManager.default_config
    _enable_plugins(enable_plugin, skip_plugin_selection, cfg)
    _configure_aws_plugin(aws_profile, cfg)
    _save_daydream_config(profile, overwrite_existing, cfg)
    mcp_server_config["mcpServers"]["daydream"] = daydream_server_config
    rich.print(
        f"Writing the following MCP Server config to Claude Desktop at: [bright_black]{config_filename}[/bright_black]"
    )
    rich.print_json(data=mcp_server_config)
    config_filename.write_text(json.dumps(mcp_server_config, indent=4))
    rich.print(
        "[green]Successfully configured Claude Desktop to use the Daydream MCP Server![/green]"
    )
    rich.print_json(data=mcp_server_config)
    if typer.confirm(
        "Open Claude Desktop now?",
        default=True,
        show_default=True,
    ):
        typer.launch(str(claude_app))


def _save_daydream_config(profile: str, overwrite_existing: bool, cfg: Config) -> None:
    cfg_path = get_config_path(profile, create=True)
    if cfg_path.exists():
        rich.print(f"Found existing config file at: [bright_black]{cfg_path}[/bright_black]")
        if not overwrite_existing:
            typer.confirm(
                "Would you like to overwrite the existing config?",
                default=False,
                show_default=True,
                abort=True,
            )
        backup_config_filename = cfg_path.with_suffix(
            f".yaml.bak.{datetime.now().strftime('%Y%m%d%H%M%S')}"
        )
        rich.print(
            f"Backing up existing Daydream '{profile}' profile config file to: [bright_black]{backup_config_filename}[/bright_black]"
        )
        shutil.copy(cfg_path, backup_config_filename)
    rich.print(
        f"Writing Daydream '{profile}' profile config to: [bright_black]{cfg_path}[/bright_black]"
    )
    save_config(cfg, profile, create=True)


def _configure_aws_plugin(aws_profile: list[str] | None, cfg: Config) -> None:
    if "aws" in cfg.plugins:
        aws_plugin_cfg = cfg.plugins["aws"]
        if aws_plugin_cfg.enabled:
            selected_aws_profiles = []
            if aws_profile:
                selected_aws_profiles = list(aws_profile)
            else:
                avail_aws_profile = boto3.session.Session().available_profiles
                enable_search = len(avail_aws_profile) > 10
                selected_aws_profiles = questionary.checkbox(
                    "Which AWS profiles would you like to use with the Daydream MCP Server? (leave empty to use all available profiles)",
                    choices=avail_aws_profile,
                    use_search_filter=enable_search,
                    use_jk_keys=not enable_search,
                ).ask()

            aws_plugin_cfg.settings["accounts"] = {
                profile_name: {
                    "profile": profile_name,
                }
                for profile_name in selected_aws_profiles
            }


def _enable_plugins(
    enable_plugin: list[str] | None, skip_plugin_selection: bool, cfg: Config
) -> None:
    if enable_plugin:
        for plugin_name, plugin_cfg in cfg.plugins.items():
            plugin_cfg.enabled = plugin_name in enable_plugin
    elif not skip_plugin_selection:
        possible_plugins = list(cfg.plugins.keys())
        enabled_search = len(possible_plugins) > 10
        selected_plugins = questionary.checkbox(
            "Which plugins would you like to enable for the Daydream MCP Server? (leave empty to enable all available plugins)",
            choices=[
                questionary.Choice(p, checked=plugin_cfg.enabled)
                for p, plugin_cfg in cfg.plugins.items()
            ],
            use_search_filter=enabled_search,
            use_jk_keys=not enabled_search,
        ).ask()
        for plugin_name, plugin_cfg in cfg.plugins.items():
            plugin_cfg.enabled = plugin_name in selected_plugins


def _ensure_config_dir_and_backup_existing(
    config_filename: Path, overwrite_existing: bool
) -> dict[str, Any]:
    mcp_server_config: dict[str, Any] = {"mcpServers": {}}
    if config_filename.exists():
        existing_content = config_filename.read_text()
        mcp_server_config = json.loads(existing_content)
        existing_daydream_config = mcp_server_config.get("mcpServers", {}).get("daydream", None)
        if existing_daydream_config:
            print(f'Found existing "daydream" config for Claude Desktop ({config_filename}):')
            rich.print_json(data={"daydream": existing_daydream_config})
            if not overwrite_existing:
                typer.confirm(
                    "Would you like to overwrite the existing config?",
                    default=False,
                    show_default=True,
                    abort=True,
                )
            backup_config_filename = config_filename.with_suffix(
                f".json.bak.{datetime.now().strftime('%Y%m%d%H%M%S')}"
            )
            rich.print(
                f"Backing up existing config file to: [bright_black]{backup_config_filename}[/bright_black]"
            )
            shutil.copy(config_filename, backup_config_filename)
    else:
        rich.print(f"Creating Claude Desktop MCP config file directory at: {config_filename}")
        config_filename.parent.mkdir(parents=True, exist_ok=True)
    return mcp_server_config
